// TODO: retrofit to use Histograms class for calculations
public class OtsuThresholdingAlgorithm {
	/*
	 * Calculates the optimal threshold level for a 2-level thresholding of the input image.
	 * The performance of this algorithm is estimated to be O(1).
	 * 
	 * This implementation could be generalized to calculate the (N-1) optimal
	 * thresholds for an N-level thresholding of the input image. It should be
	 * stressed, however, that a such implementation would probably be computationally
	 * intensive (I would estimate the performance to be O(256^N)). Even a
	 * clever implementation I believe would have a performance of around
	 * O(k*(256^(N-1) + ... + 256^1) + 256) for small k, which is
	 * asymptotically equivalent to O(256^(N-1)) which isn't much better
	 * than the O(256^N) performance of the naive algorithm.
	 */
	public static int calculateOptimalThreshold(int width, int height, int[] inGrayPixels) {
		int numPixels = inGrayPixels.length;
		return calculateOptimalThresholdWithHistogram(
			Histograms.calculateHistogramOfGrayLevels(inGrayPixels, null),
			numPixels);
	}
	
	// Average runtime (on Mac OS X): 2 milliseconds
	public static int calculateOptimalThresholdWithHistogram(int[] histogram, int numPixels) {
		return calculateOptimalThresholdWithProbabilityHistogram(
			Histograms.calculateProbabilityHistogramWithHistogram(histogram, numPixels, null));
	}
	
	public static int calculateOptimalThresholdWithProbabilityHistogram(
		double[] probabilityHistogram)
	{
		/* Begin Otsu calculation */
		double C1_omega, C1_mu;
		double C2_omega, C2_mu;
		
		double fullMu = 				// mean intensity of the entire image
			calculateFirstMoment(
				probabilityHistogram,
				0, probabilityHistogram.length);
		
		/*
		 * Find the threshold value t such that the between-class variance
		 * (between class C1=[0,t] and C2=[t+1,<max>]) is maximized.
		 * 
		 * This is the optimal value t* for t.
		 */
		double maxSigmaSquared = 0;
		int thresholdWithMaxSigmaSquared = 0;
		for (int curThreshold=0; curThreshold<256; curThreshold++) {
			C1_omega =
				calculateMass(
					probabilityHistogram,
					0, curThreshold);
			C1_mu =
				calculateCenterOfMassWithMass(
					probabilityHistogram,
					0, curThreshold,
					C1_omega);
			
			C2_omega =
				calculateMass(
					probabilityHistogram,
					curThreshold+1, probabilityHistogram.length);
			C2_mu =
				calculateCenterOfMassWithMass(
					probabilityHistogram,
					curThreshold+1, probabilityHistogram.length,
					C2_omega);
			
			// between-class variance (between class C1 and C2)
			double curSigmaSquared = 
				C1_omega*square(C1_mu - fullMu) +
				C2_omega*square(C2_mu - fullMu);
			
			if (curSigmaSquared > maxSigmaSquared) {
				maxSigmaSquared = curSigmaSquared;
				thresholdWithMaxSigmaSquared = curThreshold;
			}
		}
		
		return thresholdWithMaxSigmaSquared;
	}
	
	/* 
	 * EXPERIMENTAL - numerical stability could use some improvement
	 * 
	 * This is modified version of the Otsu algorithm that performs
	 * all computations using integers, avoiding the use of floating point numbers.
	 * 
	 * When translating the Otsu algorithm into C code, avoidance
	 * of floating point numbers *might* yield better performance.
	 */
	public static int calculateOptimalThresholdWithHistogram_noFloatingPoint(int[] histogram) {
		int C1_omega, C1_mu;
		int C2_omega, C2_mu;
		
		int image_mu =					// mean intensity of the entire image
			Histograms.getMean(
				histogram,
				0, histogram.length);
		
		long maxSigmaSquared = 0;
		int thresholdWithMaxSigmaSquared = 0;
		for (int curThreshold=0; curThreshold<256; curThreshold++) {
			C1_omega = 
				Histograms.getNumSamples(histogram, 0, curThreshold+1);
			C1_mu = 
				Histograms.getMean(
					Histograms.getSum(histogram, 0, curThreshold+1),
					C1_omega);
			
			C2_omega = 
				Histograms.getNumSamples(histogram, curThreshold+1, 256);
			C2_mu = 
				Histograms.getMean(
					Histograms.getSum(histogram, curThreshold+1, 256),
					C2_omega);
			
			// between-class variance (between class C1 and C2)
			long curSigmaSquared =
				((long) C1_omega)*((long) square(C1_mu - image_mu)) +
				((long) C2_omega)*((long) square(C2_mu - image_mu));
			
			if (curSigmaSquared > maxSigmaSquared) {
				maxSigmaSquared = curSigmaSquared;
				thresholdWithMaxSigmaSquared = curThreshold;
			}
		}
		
		return thresholdWithMaxSigmaSquared;
	}
	
	// ### MASS, FIRST MOMENT, AND CENTER OF MASS CALCULATIONS ###
	
	// TODO: retrofit to use Histograms class for calculations
	// calculate mass AKA mean level
	private static double calculateMass(
		double[] probabilityHistogram, int startIndex, int limitIndex)
	{
		double mass = 0;
		for (int i=startIndex; i<limitIndex; i++) {
			mass += probabilityHistogram[i];
		}
		return mass;
	}
	
	// TODO: retrofit to use Histograms class for calculations
	// calculate first moment AKA mean intensity
	private static double calculateFirstMoment(
		double[] probabilityHistogram, int startIndex, int limitIndex)
	{
		double firstMoment = 0;
		for (int i=startIndex; i<limitIndex; i++) {
			firstMoment += i*probabilityHistogram[i];
		}
		return firstMoment;
	}
	
	// TODO: retrofit to use Histograms class for calculations
	private static double calculateCenterOfMassWithMass(
		double[] probabilityHistogram, int startIndex, int limitIndex, double mass)
	{
		return calculateCenterOfMassWithMassAndFirstMoment(
			mass,
			calculateFirstMoment(probabilityHistogram, startIndex, limitIndex));
	}
	
	// TODO: retrofit to use Histograms class for calculations
	private static double calculateCenterOfMassWithMassAndFirstMoment(
		double mass, double firstMoment)
	{
		return (mass != 0)
			? (firstMoment / mass)
			: 0;					// center of mass is undefined
	}
	
	// ### MISC ###
	
	private static double square(double n) {
		return n*n;
	}
	
	private static int square(int n) {
		return n*n;
	}
}