import java.awt.image.BufferedImage;
import java.util.List;
import java.awt.Graphics;
import java.awt.Color;
import java.awt.Rectangle;
import java.awt.image.ImageObserver;

public class BarrelBlobDeleter implements PixelBufferToBufferedImageFilter, ParameterizedFilter {
	/** queried to find out where the barrels are in an image */
	private BarrelBlobFinder barrelFinder;
	
	// ### INIT ###
	
	public BarrelBlobDeleter() {
		this(new BarrelBlobFinder());
	}
	
	public BarrelBlobDeleter(BarrelBlobFinder barrelFinder) {
		this.barrelFinder = barrelFinder;
	}
	
	// ### FILTER ###
	
	public void filter(int width, int height, int[] inPixels,
	                   BufferedImage inImg, BufferedImage outImg)
	{
		/* Find the barrels in the input image */
		List barrelBoundList;
		{
			this.barrelFinder.filter(width, height, inPixels, inImg, null);
			barrelBoundList = this.barrelFinder.getBarrelBoundList();
		}
		
		/* Initialized the output image to match the input image */
		Graphics g = outImg.getGraphics();
		g.drawImage(inImg, 0, 0, (ImageObserver) null);
		
		if (!barrelBoundList.isEmpty()) {
			/* 
			 * Determine the average pixel value in the input image
			 * (hopefully approximating the color of the image's background)
			 */
			int avgPixel;
			{
				int numPixels = inPixels.length;
				int sumAlpha = 0, sumRed = 0, sumGreen = 0, sumBlue = 0;
				for (int i=0; i<numPixels; i++) {
					int curPixel = inPixels[i];
					sumAlpha += Pixels.getAlpha(curPixel);
					sumRed += Pixels.getRed(curPixel);
					sumGreen += Pixels.getGreen(curPixel);
					sumBlue += Pixels.getBlue(curPixel);
				}
				
				avgPixel = Pixels.packPixel(
					sumAlpha/numPixels,
					sumRed/numPixels,
					sumGreen/numPixels,
					sumBlue/numPixels);
			}
			
			/*
			 * Fill the boundary rectangles of the barrels in the image
			 * with (hopefully) the background color of the image
			 */
			g.setColor(new Color(avgPixel, false));
			for (int i=0, n=barrelBoundList.size(); i<n; i++) {
				Rectangle curBarrelBound = (Rectangle) barrelBoundList.get(i);
				g.fillRect(
					curBarrelBound.x,
					curBarrelBound.y,
					curBarrelBound.width,
					curBarrelBound.height);
			}
		}
	}
	
	public String toString() {
		return "Barrel Blob Deleter";
	}
	
	// ### PARAMETERIZED FILTER ###
	
	public FilterParam[] getParams() {
		return this.barrelFinder.getParams();
	}
	
	// ### ACCESSORS ###
	
	public BarrelBlobFinder getBarrelFinder() {
		return this.barrelFinder;
	}
}