// TODO: implement smarter convolution on image edges
public class GaussianBlurFilter implements PixelBufferFilter2, ParameterizedFilter {
	private final IntFilterParam gaussianRadiusParam =
		new IntFilterParam("Radius", 0, Integer.MAX_VALUE, 2);
	private final DoubleFilterParam gaussianStddevParam =
		new DoubleFilterParam("Stddev", 0.00, Double.MAX_VALUE, 1.00f);
	
	private final FilterParam[] params = new FilterParam[] {
		gaussianRadiusParam,
		gaussianStddevParam
	};
	
	// ### FILTER ###
	
	private int oldRadius;
	private double oldStddev;
	private double[] kernel = null;
	
	public void filter(int width, int height, int[] inPixels, int[] outPixels) {
		int gaussianRadius = gaussianRadiusParam.getValue();
		double gaussianStddev = gaussianStddevParam.getValue();
		
		/* Calculate a kernel according to the filter parameters */
		if ((this.kernel == null) || 
		    (this.oldRadius != gaussianRadius) || 
		    (this.oldStddev != gaussianStddev))
		{
			this.kernel = calculateGaussianKernel(gaussianRadius, gaussianStddev);
			
			/*
			// DEBUG: Print kernel
			for (int y=0, off=0; y<5; y++) {
				for (int x=0; x<5; x++, off++) {
					if (x != 0)
						System.out.print(", ");
					System.out.print(kernel[x]);
				}
				System.out.println("");
			}
			*/
			
			this.oldRadius = gaussianRadius;
			this.oldStddev = gaussianStddev;
		}
		
		/* Apply the Gaussian kernel to the input image */
		int kernelDiameter = gaussianRadius*2 + 1;
		applyFractionalKernel_edgesSame(
			width, height, inPixels, outPixels,
			kernelDiameter, kernelDiameter, kernel,
			gaussianRadius, gaussianRadius);
	}
	
	private static void applyFractionalKernel_edgesSame(
		int width, int height, int[] inPixels, int[] outPixels,
		int kernelWidth, int kernelHeight, double[] kernel,
		int kernelCenterX, int kernelCenterY)
	{
		/* Filter top edge */
		int off=0;
		for (int y=0; y<kernelCenterY; y++) {
			for (int x=0; x<width; x++, off++) {
				outPixels[off] = inPixels[off];
			}
		}
		
		int limitX = width-(kernelWidth-kernelCenterX);
		int limitY = height-(kernelHeight-kernelCenterY);
		int off2 = -kernelCenterX;
		for (int y=kernelCenterY; y<limitY; y++) {
			/* Filter left edge */
			for (int x=0; x<kernelCenterX; x++, off++, off2++) {
				outPixels[off] = inPixels[off];
			}
			
			/* Filter interior */
			for (int x=kernelCenterX; x<limitX; x++, off++, off2++) {
				double newAlpha = 0.0;
				double newRed = 0.0;
				double newGreen = 0.0;
				double newBlue = 0.0;
				{
					int kerOff = 0;
					for (int ky=0, rowOff=off2; ky<kernelWidth; ky++, rowOff+=width) {
						for (int kx=0, offInRow=rowOff; kx<kernelHeight; kx++, offInRow++, kerOff++) {
							int curPixel = inPixels[offInRow];
							
							newAlpha += kernel[kerOff]*Pixels.getAlpha(curPixel);
							newRed += kernel[kerOff]*Pixels.getRed(curPixel);
							newGreen += kernel[kerOff]*Pixels.getGreen(curPixel);
							newBlue += kernel[kerOff]*Pixels.getBlue(curPixel);
						}
					}
				}
				
				outPixels[off] = Pixels.packPixel(
					(int) newAlpha,
					(int) newRed,
					(int) newGreen,
					(int) newBlue);
			}
			
			/* Filter right edge */
			for (int x=limitX; x<width; x++, off++, off2++) {
				outPixels[off] = inPixels[off];
			}
		}
		
		/* Filter bottom edge */
		for (int y=limitY; y<height; y++) {
			for (int x=0; x<width; x++, off++) {
				outPixels[off] = inPixels[off];
			}
		}
	}
	
	private static double[] calculateGaussianKernel(int radius, double stddev) {
		int diameter = (radius*2)+1;
		double variance = square(stddev);
		
		/* Calculate the initial kernel */
		double[] kernel = new double[diameter*diameter];
		for (int dy=-radius, off=0; dy<=radius; dy++) {
			for (int dx=-radius; dx<=radius; dx++, off++) {
				kernel[off] = 
					StrictMath.exp(-(square(dx) + square(dy))/(2*variance)) / 
					(2*StrictMath.PI*variance);
			}
		}
		
		/* Renormalize the kernel so that its sum is 1.0 */
		{
			double sum = 0.0;
			for (int i=0, n=kernel.length; i<n; i++) {
				sum += kernel[i];
			}
			
			for (int i=0, n=kernel.length; i<n; i++) {
				kernel[i] /= sum;
			}
		}
		
		return kernel;
	}
	
	private static int square(int n) {
		return n*n;
	}
	
	private static double square(double n) {
		return n*n;
	}
	
	public String toString() {
		return "Blur (Gaussian)";
	}
	
	// ### PARAMETERIZED FILTER ###
	
	public FilterParam[] getParams() {
		return params;
	}
}