import java.util.List;
import java.util.ArrayList;
import java.awt.Image;
import java.awt.image.BufferedImage;
import java.awt.image.DataBuffer;
import java.awt.Graphics2D;
import java.awt.Color;
import java.awt.Stroke;
import java.awt.BasicStroke;
import javax.swing.*;

public class LinearHoughTransform2 implements ImageFilter, ParameterizedFilter {
	private final static boolean PRINT_HOUGH_COORDS = false;
	
	/* Values for "outputTypeParam" */
	private final static int OUTPUT_HOUGH_SPACE_RAW = 1;
	private final static int OUTPUT_HOUGH_SPACE_LOCAL_MAXIMA = 2;
	private final static int OUTPUT_HOUGH_SPACE_THRESHOLDED_LOCAL_MAXIMA = 3;
	private final static int OUTPUT_HOUGH_SPACE_BINARY_LOCAL_MAXIMA = 4;
	private final static int OUTPUT_SUPERIMPOSED_LINES = 5;
	
	// ### PARAMETERS ###
	
	// TODO: make param appear as a combo box in the GUI
	public final IntFilterParam outputTypeParam = 
		new IntFilterParam("Output Type (1-5)", 1, 5, 5);
	
	// XXX: value of parameter is UNCALIBRATED
	public final IntFilterParam normHoughValueThresholdParam = 
		new IntFilterParam("Normalized Hough-Value Threshold (Low)", 0, 255, 196);
	
	// XXX: value of parameter is UNCALIBRATED
	public final IntFilterParam windowSizeParam =
		new IntFilterParam("Local Maxima - Window Size", 1, Integer.MAX_VALUE, 10);
	
	private final FilterParam[] params = new FilterParam[] {
		outputTypeParam,
		normHoughValueThresholdParam,
		windowSizeParam
	};
	
	// ### FILTER ###
	
	private final static Stroke SUPERIMPOSED_LINE_STROKE = new BasicStroke(3.0f);	// 3 pixels wide
	
	private List maximaHoughCoords = new ArrayList();		// List<HoughPoint>
	
	/* First-time initialized */
	private static double[] sinTable = null;	// indexed by degrees; of length 360
	private static double[] cosTable = null;	// indexed by degrees; of length 360
	private static Color[] superimposedLineColors = null;
	
	/* Reinitialized upon change of image size */
	private int lastWidth = -1;
	private int lastHeight = -1;
	private int originX = -1;
	private int originY = -1;
	private int maxLineLen = -1;
	private int[] sumForLine = null;			// 2D matrix [deg,r] with the sums of the grays along the corresponding line
	private int[] sumForLine2 = null;
	private BufferedImage outHoughSpaceImage = null;
	private BufferedImage outSuperimposedImage = null;
	
	/* Every-time initialized */
	private final LocalMaximaHilighter localMaximaHilighter = new LocalMaximaHilighter();
	private BufferedImage inImgBuffer = null;
	private int[] inPixelBuffer = null;
	private int[] inGrayPixels;
	
	public Image createFilteredImage(Image img) {
		/* Load parameters */
		int outputType = outputTypeParam.getValue();
		int normHoughValueThreshold = normHoughValueThresholdParam.getValue();
		
		/* Load input image properties */
		int width = img.getWidth(null);
		int height = img.getHeight(null);
		
		/* First-time initialization */
		if (this.sinTable == null) {
			/* Compute sinTable and cosTable */
			this.sinTable = new double[360];
			this.cosTable = new double[360];
			for (int deg=0; deg<360; deg++) {
				double rad = deg*StrictMath.PI/180.0;
				this.sinTable[deg] = StrictMath.sin(rad);
				this.cosTable[deg] = StrictMath.cos(rad);
			}
			
			/* Compute superimposedLineColors as various opacities of red */
			superimposedLineColors = new Color[256];
			for (int i=0; i<256; i++) {
				superimposedLineColors[i] = new Color(255, 0, 0, i);
			}
		}
		
		/* Initialization upon change of image size */
		if ((width != lastWidth) || (height != lastHeight)) {
			this.originX = width/2;
			this.originY = height/2;
			
			this.maxLineLen = (((int) StrictMath.sqrt(width*width + height*height)) + 1)/2 + 1;		// +1's: compensate for roundoff error
			
			int sumForLine_numRs = 2*maxLineLen + 1;
			int sumForLine_numThetas = 180;
			this.sumForLine = new int[sumForLine_numRs*sumForLine_numThetas];
			this.sumForLine2 = new int[sumForLine.length];
			
			this.outHoughSpaceImage = new BufferedImage(
				sumForLine_numRs, sumForLine_numThetas,
				BufferedImage.TYPE_BYTE_GRAY);
			this.outSuperimposedImage = new BufferedImage(
				width, height,
				BufferedImage.TYPE_INT_ARGB);
			
			this.lastWidth = width;
			this.lastHeight = height;
		}
		
		/* Every-time initialization */
		this.localMaximaHilighter.windowSizeParam.setValue(windowSizeParam.getValue());
		this.inImgBuffer = BufferedImages.copyImageToARGBBufferedImage(img, this.inImgBuffer);
		this.inPixelBuffer = BufferedImages.copyBufferedImageToARGBPixelBuffer(this.inImgBuffer, this.inPixelBuffer);
		this.inGrayPixels = Pixels.ARGBPixelsToGrayPixels(inPixelBuffer, this.inGrayPixels);
		
		/* 
		 * Iterate over lines in image, summing up the gray levels along each one
		 * (incrementally calculating the value of each point in the Hough space)
		 */
		for (int thetaDeg=0, phiDeg=90, sumForLine_off=0;
		     thetaDeg<180;
		     thetaDeg++, phiDeg++)
		{
			for (int r = -maxLineLen; r<=maxLineLen; r++, sumForLine_off++) {
				double rCosThetaDeg = r*cosTable[thetaDeg];	// r*cos(thetaDeg)
				double rSinThetaDeg = r*sinTable[thetaDeg];	// r*sin(thetaDeg)
				
				/* Iterate over points in line, summing up the gray values */
				int sumOfGraysInLine = 0;
				for (int pos = -maxLineLen; pos<=maxLineLen; pos++) {
					double posCosPhiDeg = pos*cosTable[phiDeg];	// pos*cos(phiDeg)
					double posSinPhiDeg = pos*sinTable[phiDeg];	// pos*sin(phiDeg)
					
					int x = originX + (int) (rCosThetaDeg + posCosPhiDeg);	// originX + r*cos(thetaDeg) + pos*cos(phiDeg)
					int y = originY + (int) (rSinThetaDeg + posSinPhiDeg);	// originY + r*sin(thetaDeg) + pos*sin(phiDeg)
					
					if ((x >= 0) && (x < width) && (y >= 0) && (y < height)) {	// bound check
						sumOfGraysInLine += inGrayPixels[y*width + x];
					}
				}
				
				// sumForLine[thetaDeg,r] = sumOfGraysInLine
				sumForLine[sumForLine_off] = sumOfGraysInLine;
			}
		}
		
		if (outputType >= OUTPUT_HOUGH_SPACE_LOCAL_MAXIMA) {
			/* Identify local maximas in the Hough space (and eliminate the non-maximas) */
			this.localMaximaHilighter.filterGrays(
				this.outHoughSpaceImage.getWidth(),
				this.outHoughSpaceImage.getHeight(),
				sumForLine,
				sumForLine2);
			System.arraycopy(sumForLine2, 0, sumForLine, 0, sumForLine2.length);
		}
		
		/* Normalize the values in the Hough space to fall in the range [0,255] */
		{
			int maxSum = 0;
			for (int i=0, n=sumForLine.length; i<n; i++) {
				int sumForCurLine = sumForLine[i];
				if (sumForCurLine > maxSum) {
					maxSum = sumForCurLine;
				}
			}
			
			if (maxSum > 0) {
				for (int i=0, n=sumForLine.length; i<n; i++) {
					sumForLine[i] = sumForLine[i]*255/maxSum;
				}
			}
		}
		
		if (outputType >= OUTPUT_HOUGH_SPACE_THRESHOLDED_LOCAL_MAXIMA) {
			/* Threshold the normalized values in the Hough space */
			for (int i=0, n=sumForLine.length; i<n; i++) {
				if (sumForLine[i] < normHoughValueThreshold) {
					sumForLine[i] = 0;
				}
			}
		}
		
		if (outputType == OUTPUT_HOUGH_SPACE_BINARY_LOCAL_MAXIMA) {
			/* Mark local maximas in a binary fashion */
			for (int i=0, n=sumForLine.length; i<n; i++) {
				if (sumForLine[i] != 0) {
					sumForLine[i] = 255;
				}
			}
		}
		
		if (outputType < OUTPUT_SUPERIMPOSED_LINES) {
			/* Copy sumForLine data into the output Hough space image */
			DataBuffer outHoughSpaceImageBuffer = this.outHoughSpaceImage.getRaster().getDataBuffer();
			for (int i=0, n=sumForLine.length; i<n; i++) {
				outHoughSpaceImageBuffer.setElem(i, sumForLine[i]);
			}
			
			return this.outHoughSpaceImage;
		} else if (outputType == OUTPUT_SUPERIMPOSED_LINES) {
			/* 
			 * Find the coordinates in Hough space that are local maximas,
			 * storing them into "maximaHoughCoords"
			 */
			this.maximaHoughCoords.clear();		// recycle list
			for (int thetaDeg=0, phiDeg=90, sumForLine_off=0;
				 thetaDeg<180;
				 thetaDeg++, phiDeg++)
			{
				for (int r = -maxLineLen; r<=maxLineLen; r++, sumForLine_off++) {
					int value = sumForLine[sumForLine_off];
					if (value != 0) {
						this.maximaHoughCoords.add(new HoughPoint(r, thetaDeg, value));
					}
				}
			}
			
			/* Copy input image to the output image */
			BufferedImages.copyImageToARGBBufferedImage(this.inImgBuffer, this.outSuperimposedImage);
			
			/* Draw superimposed lines on output image */
			Graphics2D g = (Graphics2D) this.outSuperimposedImage.getGraphics();
			g.setStroke(SUPERIMPOSED_LINE_STROKE);
			for (int i=0, n=this.maximaHoughCoords.size(); i<n; i++) {
				HoughPoint curMaximaHoughCoord = (HoughPoint) this.maximaHoughCoords.get(i);
				int r = curMaximaHoughCoord.r;
				int thetaDeg = curMaximaHoughCoord.thetaDeg;
				int value = curMaximaHoughCoord.value;
				int phiDeg = thetaDeg + 90;
				
				double rCosThetaDeg = r*cosTable[thetaDeg];	// r*cos(thetaDeg)
				double rSinThetaDeg = r*sinTable[thetaDeg];	// r*sin(thetaDeg)
				
				int pos1 = -maxLineLen;
				int pos2 = maxLineLen;
				
				double pos1CosPhiDeg = pos1*cosTable[phiDeg];	// pos1*cos(phiDeg)
				double pos1SinPhiDeg = pos1*sinTable[phiDeg];	// pos1*sin(phiDeg)
				
				double pos2CosPhiDeg = pos2*cosTable[phiDeg];	// pos2*cos(phiDeg)
				double pos2SinPhiDeg = pos2*sinTable[phiDeg];	// pos2*sin(phiDeg)
				
				int x1 = originX + (int) (rCosThetaDeg + pos1CosPhiDeg);	// originX + r*cos(thetaDeg) + pos1*cos(phiDeg)
				int y1 = originY + (int) (rSinThetaDeg + pos1SinPhiDeg);	// originY + r*sin(thetaDeg) + pos1*sin(phiDeg)
				
				int x2 = originX + (int) (rCosThetaDeg + pos2CosPhiDeg);	// originX + r*cos(thetaDeg) + pos2*cos(phiDeg)
				int y2 = originY + (int) (rSinThetaDeg + pos2SinPhiDeg);	// originY + r*sin(thetaDeg) + pos2*sin(phiDeg)
				
				g.setColor(superimposedLineColors[value]);
				g.drawLine(x1, y1, x2, y2);
				
				if (PRINT_HOUGH_COORDS) {
					System.out.println("("+r+","+thetaDeg+") = "+value);
				}
			}
			
			return this.outSuperimposedImage;
		} else {		
			// should never get here
			throw new IllegalStateException("unknown outputType was specified");
		}
	}
	
	public String toString() {
		return "Linear Hough Transform #2";
	}
	
	// ### PARAMETERIZED FILTER ###
	
	public FilterParam[] getParams() {
		return params;
	}
	
	// ### ACCESSORS ###
	
	/**
	 * @return		a list of HoughPoints corresponding to positions/orientations
	 * 				of the identified lines; modifying this list is not recommended
	 */
	public List getHoughMaxima() {
		return this.maximaHoughCoords;
	}
	
	// ### HOUGH POINT ###
	
	public static class HoughPoint {
		int r;
		int thetaDeg;
		int value;
		
		public HoughPoint(int r, int thetaDeg, int value) {
			this.r = r;
			this.thetaDeg = thetaDeg;
			this.value = value;
		}
	}
}